#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>

#include <unistd.h>
#include <pthread.h>

#include <sys/time.h>

#define N_THREADS   (30)
#define N_COUNT     (10000)

// timestamp
inline uint64_t __attribute__((always_inline))
get_timestamp(void) {
    uint64_t ret;
    struct timeval tv;
    gettimeofday(&tv, NULL);
    ret = tv.tv_sec;
    ret *= 1000000;
    ret += tv.tv_usec;
    return ret;
}

// lock implementations

void
bad_lock(volatile uint32_t *lock) {
    while (*lock == 1);
    *lock = 1;
}

void
bad_unlock(volatile uint32_t *lock) {
    *lock = 0;
}

static inline uint32_t __attribute__((always_inline))
xchg(volatile uint32_t *lock, uint32_t xchg_value) {
    uint32_t result;
    __asm volatile("xchg %[lock], %[result]"
            : [lock] "+m" (*lock), [result] "=a" (result)
            : "a" (xchg_value)
            : "cc");
    return result;
}

void
xchg_lock(volatile uint32_t *lock) {
    while(xchg(lock, 1));
}

void
xchg_unlock(volatile uint32_t *lock) {
    xchg(lock, 0);
}


static inline uint32_t __attribute__((always_inline))
cmpxchg(volatile uint32_t *lock, uint32_t compare, uint32_t xchg_value) {
    uint32_t result;
    __asm volatile(
            "lock cmpxchg %[xchg_value], %[lock]"
            : "=a" (result), [lock] "+m" (*lock)
            : [xchg_value] "r" (xchg_value), "a" (compare)
            : "cc"
            );
    return result;
}

void
cmpxchg_lock(volatile uint32_t *lock) {
    while(cmpxchg(lock, 0, 1));
}


void
tts_xchg_lock(volatile uint32_t *lock) {
    while (1) {
        while(*lock == 1);
        if (xchg(lock, 1) == 0) {
            break;
        }
    }
}


void
backoff_cmpxchg_lock(volatile uint32_t *lock) {
    uint32_t backoff = 1;
    while(cmpxchg(lock, 0, 1)) {
        for (int i=0; i<backoff; ++i) {
            __asm volatile("pause");
        }
        if (backoff < 0x10000) {
            backoff <<= 1;
        }
    }
}


void
mutex_lock(volatile uint32_t *lock) {
    pthread_mutex_t *mutex_lock = (pthread_mutex_t *)lock;
    pthread_mutex_lock(mutex_lock);
}

void
mutex_unlock(volatile uint32_t *lock) {
    pthread_mutex_t *mutex_lock = (pthread_mutex_t *)lock;
    pthread_mutex_unlock(mutex_lock);
}

volatile uint32_t count;
volatile uint32_t lock = 0;

void *
count_no_lock(void *args) {
    for (int i=0; i < N_COUNT; ++i) {
        sched_yield();
        count += 1;
    }
}

void *
count_bad_lock(void *args) {
    for (int i=0; i < N_COUNT; ++i) {
        bad_lock(&lock);
        sched_yield();
        count += 1;
        bad_unlock(&lock);
    }
}

void *
count_xchg_lock(void *args) {
    for (int i=0; i < N_COUNT; ++i) {
        xchg_lock(&lock);
        sched_yield();
        count += 1;
        xchg_unlock(&lock);
    }
}

void *
count_cmpxchg_lock(void *args) {
    for (int i=0; i < N_COUNT; ++i) {
        cmpxchg_lock(&lock);
        sched_yield();
        count += 1;
        xchg_unlock(&lock);
    }
}

void *
count_tts_xchg_lock(void *args) {
    for (int i=0; i < N_COUNT; ++i) {
        tts_xchg_lock(&lock);
        sched_yield();
        count += 1;
        xchg_unlock(&lock);
    }
}

void *
count_backoff_cmpxchg_lock(void *args) {
    for (int i=0; i < N_COUNT; ++i) {
        backoff_cmpxchg_lock(&lock);
        sched_yield();
        count += 1;
        xchg_unlock(&lock);
    }
}

void *
count_mutex_lock(void *args) {
    for (int i=0; i < N_COUNT; ++i) {
        mutex_lock(&lock);
        sched_yield();
        count += 1;
        mutex_unlock(&lock);
    }
}




void
fork_thread(void * (*thread_func)(void*), const char *name) {
    pthread_t threads[N_THREADS];
    uint64_t time_start, time_end;

    printf("Counting %d with %d threads using %s...\n",
            N_COUNT, N_THREADS, name);

    count = 0;
    time_start = get_timestamp();

    for (int i=0; i<N_THREADS; ++i) {
        pthread_create(&threads[i], NULL, thread_func, NULL);
    }

    for (int i=0; i<N_THREADS; ++i) {
        pthread_join(threads[i], NULL);
    }

    time_end = get_timestamp();

    double elapsed_in_ms = (time_end - time_start) / 1000.0;

    printf("Count: %d, elapsed Time: %10.3f ms\n", count, elapsed_in_ms);
}

int
main(int argc, char **argv, char **envp) {
    if (argc == 1) {
        fork_thread(count_no_lock, "NO_LOCK");
        fork_thread(count_bad_lock, "BAD_LOCK");
        fork_thread(count_xchg_lock, "XCHG_LOCK");
        fork_thread(count_cmpxchg_lock, "CMPXCHG_LOCK");
        fork_thread(count_tts_xchg_lock, "TTS_LOCK");
        fork_thread(count_backoff_cmpxchg_lock, "BACKOFF_LOCK");
        fork_thread(count_mutex_lock, "MUTEX_LOCK");
    }
    else {
        for (int i=1; i<argc; ++i) {
            if (strcmp(argv[i], "no") == 0) {
                fork_thread(count_no_lock, "NO_LOCK");
            }
            else if (strcmp(argv[i], "bad") == 0) {
                fork_thread(count_bad_lock, "BAD_LOCK");
            }
            else if (strcmp(argv[i], "xchg") == 0) {
                fork_thread(count_xchg_lock, "XCHG_LOCK");
            }
            else if (strcmp(argv[i], "cmpxchg") == 0) {
                fork_thread(count_cmpxchg_lock, "CMPXCHG_LOCK");
            }
            else if (strcmp(argv[i], "tts") == 0) {
                fork_thread(count_tts_xchg_lock, "TTS_LOCK");
            }
            else if (strcmp(argv[i], "backoff") == 0) {
                fork_thread(count_backoff_cmpxchg_lock, "BACKOFF_LOCK");
            }
            else if (strcmp(argv[i], "mutex") == 0) {
                fork_thread(count_mutex_lock, "MUTEX_LOCK");
            }
        }
    }
    return 0;
}
